# -*- coding: utf-8 -*-


class Row(object):
    """

    """

    def __init__(self, row=None):
        """

        :param row: Row to copy data
        """
        self._data = {}
        if row is None:
            self._modified_data = {}
        elif isinstance(row, Row):
            self._modified_data = row.get_data()

    def __getitem__(self, field):
        """

        :param field: field name
        :type field: str
        :return: value of item oder None
        """
        return self._data.get(field, None)

    def __setitem__(self, field, value):
        """

        :param field: field name
        :type field: str
        :param value:
        :return: None
        """
        self._data[field] = value

    def get_data(self):
        """

        :return: intern data
        :rtype: dict
        """
        return self._data



# -*- coding: utf-8 -*-


from .models import Environment


class Odoo(object):
    def __init__(self, url, db, username, password, transport=None, encoding=None, verbose=0, allow_none=0,
                 use_datetime=0, context=None, use_ssl=False):
        """

        :param url:
        :param db:
        :param username:
        :param password:
        :param transport:
        :param encoding:
        :param verbose:
        :param allow_none:
        :param use_datetime:
        :param context:
        """
        self.url = url
        self.db = db
        self.username = username
        self.password = password
        self.transport = transport
        self.encoding = encoding
        self.verbose = verbose
        self.allow_none = allow_none
        self.use_datetime = use_datetime
        self.context = context
        self.use_ssl = use_ssl
        self._env = Environment(self, context=context)

    @property
    def env(self):
        """

        :return:
        :rtype: Environment
        """
        return self._env



# -*- coding: utf-8 -*-


EQUAL = '='
NOT_EQUAL = '!='

GREATER = '>'

GREATER_OR_EQUAL = '>='

LESS = '<'

LESS_OR_EQUAL = '<='

""" returns true if value is either None or False, otherwise behaves
    like =
    """
UNSET_OR_EQUAL = '=?'

LIKE_EQUAL = '=like'
""" matches field_name against the value pattern. An underscore _ in
the pattern stands for (matches) any single character; a percent
sign % matches any string of zero or more characters.
"""

""" matches field_name against the %value% pattern. Similar to =like
but wraps value with '%' before matching
"""
LIKE = 'like'

""" doesn't match against the %value% pattern
"""
NOT_LIKE = 'not like'

""" case insensitive like
"""
ILIKE = 'ilike'

""" case insensitive not like
"""
NOT_ILIKE = 'not ilike'

""" case insensitive =like
"""
ILIKE_EQUAL = '=ilike'

""" is equal to any of the items from value, value should be a list
of items
"""
IN = 'in'

""" is unequal to all of the items from value
"""
NOT_IN = 'not in'

""" is a child (descendant) of a value record.
Takes the semantics of the model into account (i.e following the
relationship field named by _parent_name).
"""
CHILD_OF = 'child_of'



# -*- coding: utf-8 -*-
"""
XmlRpc API Client for Odoo 8, 9 (not tested for odoo 10)

"""

import ssl
from socket import error as SocketError
from threading import RLock

try:
    import xmlrpclib as xmlrpc_client
except ImportError:
    from xmlrpc import client as xmlrpc_client

_Null = object()


class AuthException(Exception):
    """ Authentication Exception
    """
    pass


"""
class ConnectionPool(object):
    def __init__(self, maxsize=5, timeout=0, odoo):
        super(ConnectionPool, self).__init__()
        self._maxsize = maxsize
        self._timeout = timeout
        self._connections = Queue(maxsize=maxsize)
        self.lock = RLock()

    def get(self):
        with self.lock:
            if self._connections.qsize() < self._maxsize:
                connection = Connection(url, db, username, password, transport=None, encoding=None, verbose=0,
                                        allow_none=0,
                                        use_datetime=False, context=None, use_ssl=False)
                self._connections.put(connection)
            connection = self._connections.get(timeout=self._timeout)
            self._connections.put(connection)
"""


class Connection(object):
    """ Connect with odoo over XML-RPC protocol

    """

    def __init__(self, odoo, context={}):
        """
        Init
        :param str url: URL of Odoo Server
        :param str db: data base name
        :param str username: login username
        :param str password: login password
        :param transport:
        :param encoding:
        :param verbose:
        :param allow_none:
        :param use_datetime:
        :param context:
        :param use_ssl: Use SSL connection or not
        :type use_ssl: bool
        """
        super(Connection, self).__init__()
        self._context = context

        # Check url
        self._url = self.check_url(odoo.url)

        # SSL connection
        if odoo.use_ssl:
            context = ssl._create_unverified_context()

        # key words arguments for ServerProxy
        kwargs = {'transport': odoo.transport, 'encoding': odoo.encoding, 'verbose': odoo.verbose,
                  'allow_none': odoo.allow_none, 'use_datetime': odoo.use_datetime, 'context': context}

        # create proxies
        try:
            self._common = xmlrpc_client.ServerProxy("{}/xmlrpc/2/common".format(self._url), **kwargs)
            self._models = xmlrpc_client.ServerProxy('{}/xmlrpc/2/object'.format(self._url), **kwargs)
            self._report = xmlrpc_client.ServerProxy('{}/xmlrpc/2/report'.format(self._url))
        except xmlrpc_client.ProtocolError:
            # false url
            raise
        except SocketError:
            # Error: Odoo server unavailable
            # Internet problem
            raise

        self._db = odoo.db
        self._username = odoo.username
        self._password = odoo.password
        self._uid = None

        self.lock = RLock()

    @staticmethod
    def check_url(url):
        """
        check URL
        :param url: URL to check
        :return: corrected URL
        """
        if "http" not in url:
            raise ValueError("URL must contain http or https")
        if url[-1] == '/':
            url = url[:-1]
        else:
            url = url
        return url

    def is_locked(self):
        return self._locked

    @property
    def is_logged(self):
        """ is logged in server?
        """
        if not self._uid:
            return False
        return True

    def login(self, username=None, password=None):
        """
            Set new username, password and login
        :param username:
        :param password:
        :return:
        """
        self._username = username if username is not None else self._username
        self._password = password if password is not None else self._password
        self._login()

    def _login(self):
        """
            Login to get uid (use saved username and password)
        """
        with self.lock:
            try:
                login_method = getattr(self._common, 'authenticate', 'login')
                self._uid = login_method(self._db, self._username, self._password, {})
            except SocketError:
                # Error: Odoo server unavailable
                # Internet problem
                raise
            except Exception:
                # Database not exist
                # Username not exist
                # Password false
                raise
            if not self.is_logged:
                raise AuthException("Username or password wrong")

    def _execute_kw(self, model, method, args, kwargs):
        """ execute a method on a model

        :param str model: model name
        :param str method: method name
        :param list args: parameter as list
        :param dict kwargs: parameter as dict
        :return: response from odoo server
        """
        if not self.is_logged:
            self._login()
        kwargs['context'] = self._context
        return self._models.execute_kw(self._db, self._uid, self._password,
                                       model, method, args, kwargs)

    def execute_kw(self, model, method, *args, **kwargs):
        """

        :param model:
        :param method:
        :param args:
        :param kwargs:
        :return:
        """
        with self.lock:
            return self._execute_kw(model, method, list(args), kwargs)

    def search(self, model_name, search_domain=None, offset=0, limit=None, order=None):
        """ search() takes a mandatory domain filter (possibly empty), and returns the database identifiers of all
        records matching the filter.
        :param model_name: name of model to search
        :type model_name: str
        :param search_domain: A search domain. Use an empty list to match all records : [('id', '>', 5), (...), ...]
        :type search_domain: list
        :param offset: number of results to ignore (default: none)
        :type offset: int
        :param limit: maximum number of records to return (default: all)
        :type limit: int
        :param order: sort string
        :type order: str
        :returns: at most limit records matching the search criteria
        :rtype: list
        """
        attrs = {}
        if search_domain is None:
            search_domain = []
        if limit is not None:
            attrs['limit'] = limit
        if order is not None:
            attrs['order'] = order
        return self.execute_kw(model_name, "search", search_domain, offset=offset, **attrs)

    def search_count(self, model_name, search_domain=None):
        """Returns the number of records in the current model matching the provided domain.
        :param: model_name: name of model to search
        :type model_name: str
        :param search_domain: A search domain.
        :type search_domain: list
        :return: number of records in the current model matching the provided domain.
        :rtype: int
        """
        if search_domain is None:
            search_domain = []
        return self.execute_kw(model_name, "search_count", search_domain)

    def read(self, model_name, ids, fields=None):
        """

        :param str model_name: model name
        :param ids: ids of the rows to read
        :param list fields:
        :return:
        :rtype: list
        """
        if fields is None:
            fields = []
        return self.execute_kw(model_name, "read", ids, fields=fields)

    def fields_get(self, model_name, fields=None, attributes=None):
        """ Return the definition of each field.

        The _inherits'd fields are included. The string, help, and selection (if present) attributes are translated.
        list of description attributes to return for each field, all if empty or not provided

        :param model_name: name of model
        :type model_name: str
        :param fields: List of fields, [] for all fields
        :type fields: list
        :param string: the field's label
        :type string: bool
        :param help: a help text if available
        :type help: bool
        :param type_: to know which values to expect, or to send when updating a record
        :type type_: bool
        :return: The returned value is a dictionary (indiced by field name) of dictionaries.
        :rtype: dict
        """
        if fields is None:
            fields = []
        if attributes is None:
            attributes = ['string', 'help', 'type']
        return self.execute_kw(model_name, "fields_get", fields, attributes=attributes)

    def search_read(self, model_name, search_domain=None, offset=0, limit=None, order=None, fields=None):
        """ search() takes a mandatory domain filter (possibly empty), and returns the database identifiers of all
        records matching the filter.
        :param model_name: name of model to search
        :type model_name: str
        :param search_domain: A search domain. Use an empty list to match all records : [('id', '>', 5), (...), ...]
        :type search_domain: list
        :param offset: number of results to ignore (default: none)
        :type offset: int
        :param limit: maximum number of records to return (default: all)
        :type limit: int
        :param order: sort string
        :type order: str
        :param count: if True, only counts and returns the number of matching records (default: False)
        :type count: int
        :param fields: list of field names to return (default is all fields)
        :returns: at most limit records matching the search criteria
        :rtype: list
        """
        if search_domain is None:
            search_domain = []
        if fields is None:
            fields = []
        attrs = {}
        if search_domain is None:
            search_domain = []
        if limit is not None:
            attrs['limit'] = limit
        if order is not None:
            attrs['order'] = order
        return self.execute_kw(model_name, "search_read", search_domain, offset=offset, fields=fields, **attrs)

    def create(self, model_name, **kwargs):
        """ create a single record and return its database identifier
        :param str model_name: name of model
        :param vals: a mapping of fields to values
        :return: id of created record
        :rtype: int
        """

        # TODO: test if this method can create more than 1 record
        return self.execute_kw(model_name, "create", kwargs)

    def write(self, model_name, *ids, **vals):
        """ Records can be updated using write(), it takes a list of records to update and a mapping of updated fields
            to values similar to create()
        :param str model_name: name of model
        :param tuple ids: ids to write
        :param dict vals: values to write
        :return:
        """
        return self.execute_kw(model_name, "write", list(ids), vals)

    def unlink(self, model_name, *ids):
        """ delete records with id in the ids

        :param model_name: model name
        :param ids: tuple of ids to delete
        :rtype:
        :return:
        """
        return self.execute_kw(model_name, "unlink", list(ids))

    def check_access_rights(self, model_name, read=False, write=False, create=False, unlink=False,
                            raise_exception=False):
        """

        :param model_name:
        :param read:
        :param write:
        :param create:
        :param unlink:
        :param raise_exception:
        :return:
        """
        attrs = []
        if read:
            attrs.append('read')
        if write:
            attrs.append('write')
        if create:
            attrs.append('create')
        if unlink:
            attrs.append('unlink')
        return self.execute_kw(model_name, *attrs, raise_exception=raise_exception)

    def exec_workflow(self, model_name, workflow_name, *args):
        """ Workflows can be moved along by sending them signals.
        Instead of using the top-level execute_kw, signals are sent using exec_workflow.

        Signals are sent to a specific record, and possibly trigger a transition on the workflow instance
        associated with the record.
        :param model_name: name of model
        :param workflow_name: name of workflow to execute
        :param args: parameters for the workflow
        :return:
        """

        return self._models.exec_workflow(self._db, self._uid, self._password, model_name, workflow_name, *args)

    def render_report(self, report_name, ids, raw=False):
        """

        :param report_name: name of report
        :param ids: ids of objects to get report from
        :param raw:
        :return:
        """
        result = self._report.render_report(self._db, self._uid, self._password, report_name, ids)
        if raw:
            return result
        return result['result'].decode('base64')



# -*- coding: utf-8 -*-

from .field import Field
from exceptions import TypeError


class Criterion(object):
    def __init__(self, field_name, operator, value):
        self.field_name = field_name
        self.operator = operator
        self.value = value

    def __and__(self, other):
        if isinstance(other, Criterion):
            return Domain('&', self, other)
        else:
            raise TypeError('Object must be a Criterion')

    def __or__(self, other):
        if isinstance(other, Criterion):
            return Domain('|', self, other)
        else:
            raise TypeError('Object must be a Criterion')

    def NOT(self):
        return Domain('|', self)

    def __str__(self):
        return str(self.dump())

    def dump(self):
        return self.field_name, self.operator, self.value


class Domain:
    def __init__(self, *args):
        self.domain = []
        for arg in args:
            if isinstance(arg, Criterion):
                self.domain.append(arg)
            elif isinstance(arg, (tuple, list)):
                for mini_arg in arg:
                    self.domain.append(mini_arg)
            else:
                pass

    def __add__(self, other):
        if isinstance(other, Domain):
            pass
        elif isinstance(other, Criterion):
            pass
        else:
            pass

    def __or__(self, other):
        if isinstance(other, Domain):
            pass
        elif isinstance(other, Criterion):
            pass
        else:
            pass

    def __str__(self):
        return str(self.dump())

    def dump(self):
        res = []
        for criterion in self.domain:
            if isinstance(criterion, Criterion):
                res.append(criterion.dump())
            elif isinstance(criterion, basestring):
                res.append(criterion)
        return res



# -*- coding: utf-8 -*-

from .odoo_model import OdooModel


class Product(OdooModel):
    @property
    def __model(self):
        return "product.product"

    @property
    def __attributes(self):
        return {'name': str}


class ProductTemplate(object):
    @property
    def __model(self):
        return "product.template"

    @property
    def __attributes(self):
        return {'name': str}



# -*- coding: utf-8 -*-


class Model(object):
    @property
    def name(self):
        return self._name

    def __init__(self, name, conn):
        """

        :param name: name of data model
        :param odoo: odoo connector core
        """
        if not isinstance(name, str) or not name:
            raise TypeError("Name of model invalid")
        self._name = name
        self._conn = conn

    def search(self, search_domain=None, offset=0, limit=None, order=None):
        """ search() takes a mandatory domain filter (possibly empty), and returns the database identifiers of all
        records matching the filter.
        :param search_domain: A search domain. Use an empty list to match all records : [('id', '>', 5), (...), ...]
        :type search_domain: list
        :param offset: number of results to ignore (default: none)
        :type offset: int
        :param limit: maximum number of records to return (default: all)
        :type limit: int
        :param order: sort string
        :type order: str
        :returns: at most limit records matching the search criteria
        :rtype: list
        """
        return self._conn.search(self._name, search_domain=search_domain, offset=offset, limit=limit, order=order)

    def search_count(self, search_domain):
        """ Returns the number of records in the current model matching the provided domain.
        :param search_domain: A search domain.
        :type search_domain: list
        :return: number of records in the current model matching the provided domain.
        :rtype: int
        """
        return self._conn.search_count(self._name, search_domain)

    def read(self, ids, fields=None):
        """ Reads the requested fields for the records

        :param ids: a list of ids
        :type ids: list
        :param fields: list of field names to return (default is all fields)
        :return: a list of dictionaries mapping field names to their values, with one dictionary per record
        :rtype: list
        """
        return self._conn.read(self._name, ids, fields=fields)

    def fields_get(self, fields=None, attributes=None):
        """ Return the definition of each field.

        The _inherits'd fields are included. The string, help, and selection (if present) attributes are translated.
        list of description attributes to return for each field, all if empty or not provided

        :param fields: List of fields, [] for all fields
        :type fields: list
        :param attributes: field attribtutes to show
        :type attributes: list
        :return: The returned value is a dictionary (indiced by field name) of dictionaries.
        :rtype: dict
        """
        if attributes is None:
            attributes = ['string', 'help', 'type']
        return self._conn.fields_get(self._name, fields=fields, attributes=attributes)

    def search_read(self, search_domain=None, offset=0, limit=None, order=None, fields=None):
        """ search() takes a mandatory domain filter (possibly empty), and returns the database identifiers of all
        records matching the filter.
        :param search_domain: A search domain. Use an empty list to match all records : [('id', '>', 5), (...), ...]
        :type search_domain: list
        :param offset: number of results to ignore (default: none)
        :type offset: int
        :param limit: maximum number of records to return (default: all)
        :type limit: int
        :param order: sort string
        :type order: str
        :param fields: list of field names to return (default is all fields)
        :returns: at most limit records matching the search criteria
        :rtype: list
        """
        return self._conn.search_read(self._name, search_domain=search_domain, offset=offset, limit=limit,
                                      order=order, fields=fields)

    def create(self, **vals):
        """ The method will create a single record and return its database identifier
        :param vals: a mapping of fields to values
        :return: id of created record
        :rtype: int
        """
        return self._conn.create(self._name, **vals)

    def write(self, *ids, **vals):
        """ Records can be updated using write(), it takes a list of records to update and a mapping of updated fields
            to values similar to create()
        :param tuple ids: ids to write
        :param dict vals: values to write
        :return:
        """
        return self._conn.write(self._name, *ids, **vals)

    def unlink(self, *ids):
        """ Records can be deleted in bulk by providing their ids to unlink()
        :param ids: ids to delete
        :return:
        """
        return self._conn.unlink(self._name, *ids)

    def check_access_rights(self, model_name, read=False, write=False, create=False, unlink=False,
                            raise_exception=False):
        """

        :param model_name:
        :param read:
        :param write:
        :param create:
        :param unlink:
        :param raise_exception:
        :return:
        """
        return self._conn.check_access_rights(model_name, read=read, write=write, create=create, unlink=unlink,
                                              raise_exception=raise_exception)

    def exec_workflow(self, workflow_name, *args):
        """ Workflows can be moved along by sending them signals. 
        Instead of using the top-level execute_kw, signals are sent using exec_workflow.

        Signals are sent to a specific record, and possibly trigger a transition on the workflow instance 
        associated with the record.        
        :param workflow_name: name of workflow to execute
        :param args: parameters for the workflow
        :return:
        """
        return self._conn.exec_workflow(self._name, workflow_name, *args)

    def render_report(self, report_name, ids, raw=False):
        """

        :param report_name: name of report
        :param ids: ids of objects to get report from
        :param raw:
        :return:
        """
        return self._conn.render_report(report_name, ids, raw=raw)



# -*- coding: utf-8 -*-

from setuptools import setup, find_packages

__author__ = 'Viet Pham'
__copyright__ = 'Viet Pham'
__date__ = '27.04.2019'
__version__ = '0.2.7'
__status__ = 'beta'

setup(
    name='pydoo',
    packages=find_packages(exclude=['contrib', 'docs', 'tests']),
    version=__version__,
    description='High level library for Odoo XMLRPC API',
    author=__author__,
    author_email='dev@qvpham.de',
    url='https://github.com/julivico/pydoo',
    download_url='https://github.com/julivico/pydoo/archive/master.zip',
    keywords='python odoo xmlrpc',
    classifiers=['Development Status :: 3 - Alpha',
                 'Intended Audience :: Developers',
                 'Natural Language :: English',
                 'License :: OSI Approved :: Apache Software License',
                 'Programming Language :: Python',
                 'Programming Language :: Python :: 2.7',
                 'Programming Language :: Python :: 3',
                 'Programming Language :: Python :: 3.3',
                 'Programming Language :: Python :: 3.4',
                 'Programming Language :: Python :: 3.5',
                 'Programming Language :: Python :: 3.6',
                 'Programming Language :: Python :: 3.7',
                 'Programming Language :: Python :: Implementation :: CPython',
                 'Programming Language :: Python :: Implementation :: PyPy'],
)



# -*- coding: utf-8 -*-

from abc import ABCMeta, abstractmethod, abstractproperty


class OdooModel(object):
    __metaclass__ = ABCMeta
    __data = {}

    @abstractproperty
    def __model(self):
        return ""

    @abstractproperty
    def __attributes(self):
        return {}

    def __init__(self):
        pass



# -*- coding: utf-8 -*-

from .model import Model
from ..core import Connection


class Environment(object):
    def __init__(self, odoo, context={}):
        super(Environment, self).__init__()
        self._queue = {}
        self._odoo = odoo
        self.conn = Connection(odoo, context=context)

    @property
    def odoo(self):
        return self._odoo

    def __getitem__(self, name):
        return self._queue.setdefault(name, Model(name, self.conn))
